/*
 * Add date: {creation, modification, modified} and author: {name, email}
 * to metadata if not already present
 */

const nodegit = require('nodegit')
const path = require('path')
const moment = require('moment')

// Returns promise for patch of single commit
const gitShow = (repo, commitData) => {
  let commit
  let tree
  let parentTree
  return nodegit.Revparse.single(repo, commitData.commit)
    .then(obj => nodegit.Commit.lookup(repo, obj.id()))
    .then(_commit => {
      commit = _commit
      if (commit.parentcount() < 1) return null
      return commit.getParents(1)
    })
    .then(parents => {
      if (parents === null) return null
      return parents[0].getTree()
    })
    .then(_parentTree => {
      parentTree = _parentTree
      return commit.getTree()
    })
    .then(_tree => {
      tree = _tree
      return nodegit.Diff.treeToTree(repo, parentTree, tree, null)
    })
    .then(diff => diff.patches())
    .then(patches => [patches, commitData])
}

const gitlog = repo => {
  const map = () =>
    Promise.resolve().then(() => commit => {
      const author = commit.author()
      const sha = commit.sha()

      return {
        commit: sha,
        author: {
          name: author.name(),
          email: author.email(),
        },
        date: commit.date(),
        message: commit.message(),
      }
    })

  return repo.getHeadCommit().then(commit =>
    new Promise((resolve, reject) => {
      commit
        .history()
        .on('end', resolve)
        .on('error', reject)
        .start()
    }).then(commits => map().then(log => commits.map(log))),
  )
}

const getFilesHistory = (repo, _startCommitMessage) => {
  let startCommitMessage = _startCommitMessage
  if (!Array.isArray(startCommitMessage) && startCommitMessage !== undefined) {
    startCommitMessage = [startCommitMessage]
  }
  return Promise.resolve()
    .then(() => gitlog(repo))
    .then(log => {
      const promises = []
      for (let i = 0; i < log.length; i += 1) {
        promises.push(gitShow(repo, log[i]))
      }
      return Promise.all(promises)
    })
    .then(diffs => {
      const filehistory = {}
      diffs.forEach(data => {
        const commit = data[1]
        const diff = data[0]

        diff.forEach(entry => {
          const p = path.normalize(entry.newFile().path())
          if (filehistory[p] === undefined) {
            filehistory[p] = []
          }
          filehistory[p].push(commit)
        })
      })
      return filehistory
    })
    .then(log => {
      const sorter = (a, b) => {
        const rhs = a.date
        const lhs = b.date
        return lhs > rhs ? 1 : lhs < rhs ? -1 : 0
      }
      for (const file of Object.keys(log)) {
        log[file].sort(sorter)
      }
      return log
    })
    .then(log => {
      if (startCommitMessage === undefined) return log
      function mergeCommit(commits) {
        let r
        commits.reverse()
        commits.forEach(commit => {
          if (r === undefined)
            startCommitMessage.forEach(matcher => {
              if (commit.message.indexOf(matcher) > -1) r = commit
            })
        })
        commits.reverse()
        return r
      }
      for (const file of Object.keys(log)) {
        // eslint-disable-next-line no-param-reassign
        log[file].mergeCommit = mergeCommit(log[file])
      }
      return log
    })
}

module.exports = obj => {
  console.log('Build step: GitInfo')
  const pathToRepo = path.resolve(obj.config.sourceDir)
  return new Promise((resolve, reject) =>
    nodegit.Repository.open(pathToRepo)
      .catch(() => {
        resolve(obj)
      })
      .then(repo => getFilesHistory(repo, obj.config.git.mergeMessage))
      .then(hist => {
        if (obj.config.options.debugarticle !== null) {
          const log = hist[obj.config.options.debugarticle]
          log.forEach(commit => {
            console.log(
              `   ${commit.date} ${
                commit.author.email
              }\t${commit.message.replace(/\n([^]*)$/, '')}`,
            )
          })
          console.log()
          console.log(
            `   Starting commit: ${log.mergeCommit.message.replace(
              /\n([^]*)$/,
              '',
            )}`,
          )
        }
        return hist
      })
      .then(filesHistory => {
        obj.pages.forEach(article => {
          const file = path.relative(pathToRepo, article.filename)
          const commits = filesHistory[file]
          if (commits !== undefined) {
            commits.sort((a, b) => b.date - a.date)
            const newestCommit = commits[0]
            const oldestCommit = commits[commits.length - 1]
            const { mergeCommit } = commits

            // eslint-disable-next-line no-param-reassign
            if (article.metadata.date === undefined) article.metadata.date = {}

            if (article.metadata.date.creation === undefined) {
              if (mergeCommit === undefined) {
                console.log(
                  `[Warning] Article ${file} does not have merge commit.`,
                )
                // eslint-disable-next-line no-param-reassign
                article.metadata.date.creation = moment()
              } else {
                // eslint-disable-next-line no-param-reassign
                article.metadata.date.creation = mergeCommit.date
              }
            }

            // eslint-disable-next-line no-param-reassign
            article.metadata.date.modification = newestCommit.date

            if (article.metadata.author === undefined) {
              // eslint-disable-next-line no-param-reassign
              article.metadata.author = {}
            }
            if (article.metadata.author.name === undefined) {
              // eslint-disable-next-line no-param-reassign
              article.metadata.author.name = oldestCommit.author.name
            }
            if (article.metadata.author.email === undefined) {
              // eslint-disable-next-line no-param-reassign
              article.metadata.author.email = oldestCommit.author.email
            }

            // eslint-disable-next-line no-param-reassign
            article.commits = commits
          } else {
            console.log(`[Warning] Article ${file} is not in git repository.`)
          }
        })
        return obj
      })
      .then(o => {
        resolve(o)
      })
      .catch(e => {
        reject(e)
      }),
  )
}
