const syntaxError = require('syntax-error')
const fs = require('fs')
const chokidar = require('chokidar')

const jobList = {}
let watchers = []

function JobError(message, name) {
  Error.captureStackTrace(this, JobError)
  this.name = `${name}Error`
  this.message = message
  return this
}

let onReload = () => {}
const requireError = (e, jobName, module, reject) => {
  if (e instanceof SyntaxError) {
    console.log(`${module}: ${e}`)
    const resolved = require.resolve(module)
    fs.readFile(resolved, 'utf8', (err, content) => {
      console.log(syntaxError(content, resolved))
      reject(
        new JobError(
          `Failed to register job ${jobName} (SyntaxError)`,
          'Require',
        ),
      )
    })
  } else {
    if (e.code === 'MODULE_NOT_FOUND') console.log(e.toString())
    reject(new JobError(`Failed to register job ${jobName} (${e})`, 'Require'))
  }
}

const jobs = {
  register(jobName, module, watch = true) {
    return new Promise((resolve, reject) => {
      if (jobList[jobName] !== undefined) {
        Promise.reject(
          new JobError(
            `Job ${jobName} is already registered`,
            'JobAlreadyRegistered',
          ),
        )
        return
      }

      let f
      try {
        // eslint-disable-next-line global-require, import/no-dynamic-require
        f = require(module)
      } catch (e) {
        requireError(e, jobName, module, reject)
        return
      }

      if (typeof f !== 'function')
        throw new JobError(
          `Module "${module}" for job "${jobName}" does not export function`,
        )
      jobList[jobName] = {
        f,
        module,
      }
      if (watch) jobs.watch(jobName)
      resolve()
    })
  },
  reload(jobName) {
    const { module } = jobList[jobName]
    delete require.cache[require.resolve(module)]

    return new Promise((resolve, reject) => {
      try {
        // eslint-disable-next-line global-require, import/no-dynamic-require
        const f = require(module)
        if (typeof f === 'function') {
          const job = {
            f,
            module,
          }
          delete jobList[jobName]
          jobList[jobName] = job
          resolve()
        } else {
          console.log(`Error reloading job ${jobName} retry in 500ms`)
          setTimeout(() => {
            resolve(jobs.reload(jobName))
          }, 500)
        }
      } catch (e) {
        console.log(e)
        requireError(e, jobName, module, reject)
      }
    })
  },
  // registerMultiple(
  //  {watch: true},
  //  ['jobName','./module'],['jobName2','./module2']
  // )
  registerMultiple({ watch = true }, ...args) {
    const regs = []
    for (const job of args) {
      regs.push(jobs.register(job[0], job[1], watch))
    }
    return Promise.all(regs)
  },
  run(jobName, ...args) {
    if (jobList[jobName] === undefined) {
      throw new JobError(`Job ${jobName} is not registered`, 'JobNotRegistered')
    }
    try {
      return jobList[jobName].f(...args)
    } catch (e) {
      console.log('Error running job', jobName)
      console.log(jobList[jobName])
      throw e
    }
  },
  runSequence(...args) {
    let prom
    console.log('Running', args)
    const jobSequence = args
    jobSequence.forEach(job => {
      if (prom === undefined) {
        prom = jobs.run(job)
      } else {
        prom = prom.then((...a) => {
          a.splice(0, 0, job)
          return jobs.run(...a)
        })
      }
    })
    return prom
  },
  watch(jobName) {
    const { module } = jobList[jobName]
    const file = require.resolve(module)
    const w = chokidar.watch(file)
    w.on('change', () => {
      jobs.reload(jobName).then(onReload)
    })
    watchers.push(w)
  },
  onReload(f) {
    onReload = f
  },
  close() {
    watchers.forEach(w => {
      w.close()
      console.log(w.getWatched())
    })
    watchers = []
  },
}

module.exports = jobs
module.exports.jobList = jobList
